'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

exports.transformArrayComprehension = transformArrayComprehension;
exports.transformObjectComprehension = transformObjectComprehension;

var _babelInstance = require('./babelInstance');

var _tails = require('./tails');

function validateComprehensionLoopBody(loopBodyPath) {
  return loopBodyPath.traverse({
    noScope: true,

    Function(path) {
      path.skip();
    }, AwaitExpression(awaitPath) {
      throw awaitPath.buildCodeFrameError("`await` is not allowed within Comprehensions; " + "instead, await the Comprehension (eg; `y <- [for x of xs: x]`).");
    }, YieldExpression(yieldPath) {
      throw yieldPath.buildCodeFrameError("`yield` is not allowed within Comprehensions.");
    }, ReturnStatement(returnPath) {
      throw returnPath.buildCodeFrameError("`return` is not allowed within Comprehensions.");
    } });
}

function wrapComprehensionInIife(bodyVarId, bodyVarInitializer, loopBody) {
  var t = (0, _babelInstance.getTypes)();

  return t.callExpression(t.arrowFunctionExpression([], t.blockStatement([t.variableDeclaration("const", [t.variableDeclarator(bodyVarId, bodyVarInitializer)]), loopBody, t.returnStatement(bodyVarId)])), []);
}

function transformArrayComprehension(path) {
  var t = (0, _babelInstance.getTypes)();
  validateComprehensionLoopBody(path.get("loop.body"));

  var id = path.scope.generateUidIdentifier("arr");
  (0, _tails.transformTails)(path.get("loop.body"), true, true, function (expr) {
    return t.callExpression(t.memberExpression(id, t.identifier("push")), [expr]);
  });

  return path.replaceWith(wrapComprehensionInIife(id, t.arrayExpression(), path.node.loop));
}

function transformObjectComprehension(path) {
  var t = (0, _babelInstance.getTypes)();
  validateComprehensionLoopBody(path.get("loop.body"));

  var id = path.scope.generateUidIdentifier("obj");
  (0, _tails.transformTails)(path.get("loop.body"), true, true, function (seqExpr, tailPath) {
    if (seqExpr.type !== "SequenceExpression" || seqExpr.expressions.length !== 2) {
      throw tailPath.buildCodeFrameError("Object comprehensions must end" + " with a (key, value) pair.");
    }
    var _seqExpr$expressions = _slicedToArray(seqExpr.expressions, 2),
        keyExpr = _seqExpr$expressions[0],
        valExpr = _seqExpr$expressions[1];

    return t.assignmentExpression("=", t.memberExpression(id, keyExpr, true), valExpr);
  });

  return path.replaceWith(wrapComprehensionInIife(id, t.objectExpression([]), path.node.loop));
}