'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.isDestructuringPattern = isDestructuringPattern;
exports.matchesDestructuringPattern = matchesDestructuringPattern;
exports.matchesArrayPattern = matchesArrayPattern;
exports.matchesObjectPattern = matchesObjectPattern;
exports.toRestructureExpression = toRestructureExpression;
exports.toRestructureExpressionForArrayPattern = toRestructureExpressionForArrayPattern;
exports.toRestructureExpressionForObjectPattern = toRestructureExpressionForObjectPattern;

var _lightscriptTypes = require('./lightscriptTypes');

var _lightscriptTypes2 = _interopRequireDefault(_lightscriptTypes);

var _logical = require('./logical');

var _helpers = require('./helpers');

var _is = require('./is');

var _placeAtNode = require('ast-loc-utils/lib/placeAtNode');

var _placeAtNode2 = _interopRequireDefault(_placeAtNode);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function isDestructuringPattern(node) {
  return node.type === "ArrayPattern" || node.type === "ObjectPattern";
} // Generate a LogicalExpression that asserts a ref is compatible with a
// destructuring pattern.

function matchesDestructuringPattern(ref, pattern) {
  var _ref = pattern.type;
  if (_ref === "ArrayPattern") return matchesArrayPattern(ref, pattern);else if (_ref === "ObjectPattern") return matchesObjectPattern(ref, pattern);
}

////////// Array pattern matching


function Bracket(array, i, element) {
  return (0, _placeAtNode2.default)(_lightscriptTypes2.default.memberExpression(array, (0, _placeAtNode2.default)(_lightscriptTypes2.default.numericLiteral(i), element), true), element);
}

function HasLength(array, min, max, patternNode) {
  var args = max ? [array, (0, _placeAtNode2.default)(_lightscriptTypes2.default.numericLiteral(min), patternNode), (0, _placeAtNode2.default)(_lightscriptTypes2.default.numericLiteral(max), patternNode)] : min > 0 ? [array, (0, _placeAtNode2.default)(_lightscriptTypes2.default.numericLiteral(min), patternNode)] : [array];

  return _lightscriptTypes2.default.callExpression((0, _placeAtNode2.default)(_lightscriptTypes2.default.clone((0, _helpers.helper)('hasLength')), patternNode), args);
}

function matchesArrayPattern(array, arrayPattern) {
  var elements = arrayPattern.elements;

  var minLength = 0;
  var conditions = function () {
    var _arr = [];

    for (var _i5 = 0, _len6 = elements.length; _i5 < _len6; _i5++) {
      var _element = elements[_i5];

      var elementType = _element == null ? null : _element.type;

      // Check for mandatory elements
      if (elementType !== "RestElement" && elementType !== "AssignmentPattern") {
        minLength = _i5 + 1;
      }if (_element && isDestructuringPattern(_element)) {
        _arr.push(matchesDestructuringPattern(Bracket(array, _i5, _element), _element));
      }
    }

    return _arr;
  }();

  var maxLength = !(elements.length > 0 && elements[elements.length - 1].type === "RestElement") ? elements.length : null;

  // Length conditionals
  conditions.unshift(HasLength(array, minLength, maxLength, arrayPattern));

  return (0, _logical.createAndExpression)(conditions);
} ////////// Object pattern matching


function In(key, obj) {
  return (0, _placeAtNode2.default)(_lightscriptTypes2.default.binaryExpression('in', (0, _placeAtNode2.default)(_lightscriptTypes2.default.stringLiteral(key.name), key), obj), key);
}

function Dot(obj, key) {
  return (0, _placeAtNode2.default)(_lightscriptTypes2.default.memberExpression(obj, key), key);
}

function HasProps(obj, patternNode, props) {
  var args = props.slice();
  args.unshift(obj);
  return (0, _placeAtNode2.default)(_lightscriptTypes2.default.callExpression((0, _placeAtNode2.default)(_lightscriptTypes2.default.clone((0, _helpers.helper)('hasProps')), patternNode), args), patternNode);
}

function matchesObjectPattern(obj, objectPattern) {
  var properties = objectPattern.properties;

  // _hasProps(_obj, ...)

  var conditions = [HasProps(obj, objectPattern, function () {
    var _arr2 = [];

    for (var _i6 = 0, _len7 = properties.length; _i6 < _len7; _i6++) {
      var _property = properties[_i6];

      if ((0, _is.isa)(_property, "ObjectProperty")) {
        var key = _property.key,
            value = _property.value;

        if (!(0, _is.isa)(value, "AssignmentPattern") && (0, _is.isa)(key, "Identifier")) {
          _arr2.push((0, _placeAtNode2.default)(_lightscriptTypes2.default.stringLiteral(key.name), key));
        }
      }
    }

    return _arr2;
  }())];

  // Conditions for nested patterns
  var nestedConditions = function () {
    var _arr3 = [];

    for (var _i7 = 0, _len8 = properties.length; _i7 < _len8; _i7++) {
      var _property2 = properties[_i7];

      if ((0, _is.isa)(_property2, "ObjectProperty")) {
        var key = _property2.key,
            value = _property2.value;

        if (value.type === "AssignmentPattern" && isDestructuringPattern(value.left)) {
          _arr3.push((0, _logical.createOrExpression)((0, _logical.createNotExpression)(In(key, obj)), matchesDestructuringPattern(Dot(obj, key), value.left)));
        } else if (isDestructuringPattern(value)) {
          _arr3.push(matchesDestructuringPattern(Dot(obj, key), value));
        }
      }
    }

    return _arr3;
  }();

  return (0, _logical.createAndExpression)(conditions.concat(nestedConditions));
} ////////// "Restructuring" after a destructure


function toRestructureExpression(pattern) {
  var _ref2 = pattern.type;
  if (_ref2 === "ArrayPattern") return toRestructureExpressionForArrayPattern(pattern);else if (_ref2 === "ObjectPattern") return toRestructureExpressionForObjectPattern(pattern);
}

function maybeToRestructureExpression(node) {
  if (isDestructuringPattern(node)) return toRestructureExpression(node);else return node;
}

function toRestructureExpressionForArrayPattern(pattern) {
  return (0, _placeAtNode2.default)(_lightscriptTypes2.default.arrayExpression(function () {
    var _arr4 = [];

    for (var _arr8 = pattern.elements, _i8 = 0, _len9 = _arr8.length; _i8 < _len9; _i8++) {
      var _element2 = _arr8[_i8];

      if ((0, _is.isa)(_element2, "AssignmentPattern")) _arr4.push(maybeToRestructureExpression(_element2.left));else if ((0, _is.isa)(_element2, "RestElement")) _arr4.push((0, _placeAtNode2.default)(_lightscriptTypes2.default.spreadElement(_element2.argument), _element2));else _arr4.push(maybeToRestructureExpression(_element2));
    }

    return _arr4;
  }()), pattern);
}

function restructureObjAssignment(assign, defaultKey) {
  if (isDestructuringPattern(assign.left)) {
    return _lightscriptTypes2.default.objectProperty(defaultKey, toRestructureExpression(assign.left));
  } else {
    return _lightscriptTypes2.default.objectProperty(assign.left, assign.left, false, true);
  }
}

function toRestructureExpressionForObjectPattern(pattern) {
  return (0, _placeAtNode2.default)(_lightscriptTypes2.default.objectExpression(function () {
    var _arr6 = [];

    for (var _arr9 = pattern.properties, _i9 = 0, _len10 = _arr9.length; _i9 < _len10; _i9++) {
      var _property3 = _arr9[_i9];

      if ((0, _is.isa)(_property3, "RestProperty")) {
        _arr6.push((0, _placeAtNode2.default)(_lightscriptTypes2.default.spreadProperty(_property3.argument), _property3));
      } else if (isDestructuringPattern(_property3.value)) {
        _arr6.push((0, _placeAtNode2.default)(_lightscriptTypes2.default.objectProperty(_property3.key, toRestructureExpression(_property3.value)), _property3));
      } else if ((0, _is.isa)(_property3.value, "AssignmentPattern")) {
        _arr6.push((0, _placeAtNode2.default)(restructureObjAssignment(_property3.value, _property3.key), _property3));
      } else if ((0, _is.isa)(_property3.value, "Identifier")) {
        _arr6.push((0, _placeAtNode2.default)(_lightscriptTypes2.default.objectProperty(_property3.value, _property3.value, false, true), _property3));
      }
    }

    return _arr6;
  }()), pattern);
}