"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.getInlinedOperatorsEnabled = getInlinedOperatorsEnabled;
exports.replaceWithInlinedOperator = replaceWithInlinedOperator;

var _babelInstance = require("./babelInstance");

var inlinedOperator = {
  looseEq: function looseEq(t, args) {
    return t.binaryExpression("==", args[0], args[1]);
  },
  looseNotEq: function looseNotEq(t, args) {
    return t.binaryExpression("!=", args[0], args[1]);
  },
  bitwiseNot: function bitwiseNot(t, args) {
    return t.unaryExpression("~", args[0]);
  },
  bitwiseAnd: function bitwiseAnd(t, args) {
    return t.binaryExpression("&", args[0], args[1]);
  },
  bitwiseOr: function bitwiseOr(t, args) {
    return t.binaryExpression("|", args[0], args[1]);
  },
  bitwiseXor: function bitwiseXor(t, args) {
    return t.binaryExpression("^", args[0], args[1]);
  },
  bitwiseLeftShift: function bitwiseLeftShift(t, args) {
    return t.binaryExpression("<<", args[0], args[1]);
  },
  bitwiseRightShift: function bitwiseRightShift(t, args) {
    return t.binaryExpression(">>", args[0], args[1]);
  },
  bitwiseZeroFillRightShift: function bitwiseZeroFillRightShift(t, args) {
    return t.binaryExpression(">>>", args[0], args[1]);
  }
};

function getInlinedOperatorsEnabled(opts) {
  return !(opts.stdlib === false || typeof opts.stdlib === "object" && opts.stdlib.lightscript === false);
}

function replaceWithInlinedOperator(path, callee, args) {
  var t = (0, _babelInstance.getTypes)();
  if (callee.type !== "Identifier") return false;
  var name = callee.name;
  if (!inlinedOperator[name] || path.scope.hasBinding(name)) return false;
  path.replaceWith(inlinedOperator[name](t, args));
  return true;
}