'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.transformMatchExpression = transformMatchExpression;
exports.transformMatchStatement = transformMatchStatement;

var _lightscriptTypes = require('./lightscriptTypes');

var _lightscriptTypes2 = _interopRequireDefault(_lightscriptTypes);

var _blocks = require('./blocks');

var _is = require('./is');

var _destructure = require('./destructure');

var _logical = require('./logical');

var _variables = require('./variables');

var _getLoc = require('ast-loc-utils/lib/getLoc');

var _getLoc2 = _interopRequireDefault(_getLoc);

var _placeAtLoc = require('ast-loc-utils/lib/placeAtLoc');

var _placeAtLoc2 = _interopRequireDefault(_placeAtLoc);

var _placeAtNode = require('ast-loc-utils/lib/placeAtNode');

var _placeAtNode2 = _interopRequireDefault(_placeAtNode);

var _getSurroundingLoc = require('ast-loc-utils/lib/getSurroundingLoc');

var _getSurroundingLoc2 = _interopRequireDefault(_getSurroundingLoc);

var _span = require('ast-loc-utils/lib/span');

var _span2 = _interopRequireDefault(_span);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toArray(arr) { return Array.isArray(arr) ? arr : Array.from(arr); }

function isUndefined(node) {
  return node.type === "Identifier" && node.name === "undefined";
}

function isSignedNumber(node) {
  return node.type === "UnaryExpression" && (node.operator === "+" || node.operator === "-") && node.argument.type === "NumericLiteral";
}

function looksLikeClassName(id) {
  var name = id.name;

  if (name[0].toUpperCase() === name[0]) {
    if (name.length === 1) return true;else if (name.toUpperCase() === name) return false;else return true;
  } else return false;
}

function isSimpleMemberExpression(node) {
  return (0, _is.isa)(node, "MemberExpression") && !node.computed && ((0, _is.isa)(node.object, "Identifier") || isSimpleMemberExpression(node.object));
}

function isClassMatchable(node) {
  return (0, _is.isa)(node, "Identifier") && looksLikeClassName(node) || isSimpleMemberExpression(node) && looksLikeClassName(node.property);
}

function hasTypeof(ref, typeName, loc) {
  return (0, _placeAtLoc2.default)(_lightscriptTypes2.default.binaryExpression("===", (0, _placeAtLoc2.default)(_lightscriptTypes2.default.unaryExpression("typeof", ref), loc), (0, _placeAtLoc2.default)(_lightscriptTypes2.default.stringLiteral(typeName), loc)), loc);
}

function createClassMatchCondition(node, discriminantRef) {
  var name = (0, _is.isa)(node, "MemberExpression") ? node.property.name : node.name;

  if (name === "Number") return hasTypeof(discriminantRef, "number", (0, _getLoc2.default)(node));else if (name === "String") return hasTypeof(discriminantRef, "string", (0, _getLoc2.default)(node));else if (name === "Boolean") return hasTypeof(discriminantRef, "boolean", (0, _getLoc2.default)(node));else if (name === "Array") {
    return (0, _placeAtNode2.default)(_lightscriptTypes2.default.callExpression((0, _placeAtNode2.default)(_lightscriptTypes2.default.memberExpression((0, _placeAtNode2.default)(_lightscriptTypes2.default.identifier("Array"), node), (0, _placeAtNode2.default)(_lightscriptTypes2.default.identifier("isArray"), node)), node), [discriminantRef]), node);
  } else {
    return (0, _placeAtNode2.default)(_lightscriptTypes2.default.binaryExpression("instanceof", discriminantRef, node), node);
  }
}

function createMatchTestReplacement(path, discriminantRef) {
  var node = path.node;

  if ((0, _is.isa)(node, "LogicalExpression")) {
    node.left = createMatchTestReplacement(path.get("left"), discriminantRef);
    node.right = createMatchTestReplacement(path.get("right"), discriminantRef);
    return node;
  } else if ((0, _is.isa)(node, "UnaryExpression") && node.operator === "!") {
    node.argument = createMatchTestReplacement(path.get("argument"), discriminantRef);
    return node;
  } else if ((0, _is.isa)(node, "RegExpLiteral")) {
    var loc = (0, _getLoc2.default)(node);
    return (0, _placeAtLoc2.default)(_lightscriptTypes2.default.callExpression((0, _placeAtLoc2.default)(_lightscriptTypes2.default.memberExpression(node, (0, _placeAtLoc2.default)(_lightscriptTypes2.default.identifier("test"), loc)), loc), [discriminantRef]), loc);
  } else if (isUndefined(node) || (0, _is.isa)(node, "Literal") || isSignedNumber(node)) {
    return (0, _placeAtNode2.default)(_lightscriptTypes2.default.binaryExpression("===", discriminantRef, node), node);
  } else if ((0, _is.isa)(node, "ObjectExpression") || (0, _is.isa)(node, "ArrayExpression")) {
    throw path.buildCodeFrameError("Object and array literals are illegal in test expressions. " + "To match against a destructuring pattern, " + "use `with` (eg: `| with { foo, bar }: foo + bar`).");
  } else if (isClassMatchable(node)) {
    return createClassMatchCondition(node, discriminantRef);
  } else {
    return node;
  }
}

function transformPreTest(_ref2, testPath) {
  var discriminantRef = _ref2.discriminantRef;

  return createMatchTestReplacement(testPath, discriminantRef);
}

function transformMatchTest(mtch, casePath) {
  var _casePath$node = casePath.node,
      test = _casePath$node.test,
      binding = _casePath$node.binding;

  var testPath = casePath.get("test");
  var conditions = [];
  var discriminantRef = mtch.discriminantRef;


  if (test) conditions.push(transformPreTest(mtch, testPath));
  if (binding && (0, _destructure.isDestructuringPattern)(binding)) {
    conditions.push((0, _destructure.matchesDestructuringPattern)(discriminantRef, binding));
  }return (0, _logical.createAndExpression)(conditions);
}

function transformBoundConsequent(_ref3, binding, consequent) {
  var discriminantRef = _ref3.discriminantRef;

  var block = (0, _blocks.toBlockStatement)(consequent);
  block.body.unshift((0, _placeAtNode2.default)(_lightscriptTypes2.default.variableDeclaration("const", [(0, _placeAtNode2.default)(_lightscriptTypes2.default.variableDeclarator(binding, discriminantRef), binding)]), binding));
  return block;
}

function transformMatchConsequent(mtch, casePath) {
  var _casePath$node2 = casePath.node,
      binding = _casePath$node2.binding,
      consequent = _casePath$node2.consequent;

  if (binding) {
    return transformBoundConsequent(mtch, binding, consequent);
  } else {
    return (0, _blocks.toBlockStatement)(consequent);
  }
}

function transformMatchCase(mtch, casePath) {
  var _ref;

  var c = casePath.node;
  return {
    test: ((_ref = c.test) == null ? null : _ref.type) !== "MatchElse" ? transformMatchTest(mtch, casePath) : null,
    consequent: transformMatchConsequent(mtch, casePath)
  };
}

function transformMatchCases(mtch, casePaths) {
  var _casePaths = _toArray(casePaths),
      casePath = _casePaths[0],
      rest = _casePaths.slice(1);

  var c = casePath.node;

  var _transformMatchCase = transformMatchCase(mtch, casePath),
      test = _transformMatchCase.test,
      consequent = _transformMatchCase.consequent;

  if (rest.length > 0) {
    var loc = (0, _getSurroundingLoc2.default)([c, mtch.lastCase]);
    return (0, _placeAtLoc2.default)(_lightscriptTypes2.default.ifStatement(test, consequent, transformMatchCases(mtch, rest)), loc);
  } else if (test) {
    return (0, _placeAtNode2.default)(_lightscriptTypes2.default.ifStatement(test, consequent), c);
  } else {
    return consequent;
  }
} // Computed values needed in recursive descent.


function getMatchInfo(path, discriminantRef) {
  var node = path.node;

  return { path, node, discriminantRef, lastCase: node.cases[node.cases.length - 1] };
}

function transformMatchExpression(path) {
  var node = path.node;
  var discriminant = node.discriminant,
      _node$alias = node.alias,
      alias = _node$alias === undefined ? (0, _placeAtLoc2.default)(_lightscriptTypes2.default.identifier("it"), (0, _span2.default)((0, _getLoc2.default)(node), 1)) : _node$alias;


  var matchBody = transformMatchCases(getMatchInfo(path, alias), path.get("cases"));

  var iife = (0, _placeAtNode2.default)(_lightscriptTypes2.default.callExpression((0, _placeAtNode2.default)(_lightscriptTypes2.default.arrowFunctionExpression([alias], (0, _placeAtNode2.default)(_lightscriptTypes2.default.blockStatement([matchBody]), node)), node), [discriminant]), node);

  return path.replaceWith(iife);
}

function transformMatchStatement(path) {
  var node = path.node;
  var discriminant = node.discriminant,
      _node$alias2 = node.alias,
      alias = _node$alias2 === undefined ? (0, _placeAtLoc2.default)(_lightscriptTypes2.default.identifier("it"), (0, _span2.default)((0, _getLoc2.default)(node), 1)) : _node$alias2;


  var elideAlias = (0, _is.isa)(discriminant, "Identifier") && !(0, _variables.containsReferencedIdentifier)(path, alias);
  var needsBlock = !elideAlias && path.scope.hasBinding(alias.name);

  if (elideAlias) {
    return path.replaceWith(transformMatchCases(getMatchInfo(path, discriminant), path.get("cases")));
  } else if (needsBlock) {
    return path.replaceWith((0, _placeAtNode2.default)(_lightscriptTypes2.default.blockStatement([(0, _placeAtNode2.default)(_lightscriptTypes2.default.variableDeclaration("const", [(0, _placeAtNode2.default)(_lightscriptTypes2.default.variableDeclarator(alias, discriminant), discriminant)]), discriminant), transformMatchCases(getMatchInfo(path, alias), path.get("cases"))]), node));
  } else {
    (0, _variables.insertDeclarationBefore)(path, alias, discriminant, "const", (0, _getLoc2.default)(discriminant));
    return path.replaceWith(transformMatchCases(getMatchInfo(path, alias), path.get("cases")));
  }
}